/*
 * General header file for 1616/OS V4.2b system calls
 *
 * Andrew Morton, Applix pty limited
 */

#ifndef STDIN
#define STDERR		0x100	/* Standard error file descriptor */
#define STDOUT		0x101	/* Standard output file descriptor */
#define STDIN		0x102	/* Standard input file descriptor */
#endif

/*
 * We macro all the system calls as calls to trap7().
 * The system calls are in upper case to distinguish them from
 * normal function calls, such as C library functions
 *
 * Note that the printf, sprintf and fprintf calls are missing here
 * because they use a variable number of arguments, which confuses C
 * preprocessors.
 */

extern long trap7(int callnr, ...);

#ifndef TRAP7
#define TRAP7 trap7
#endif

#define COLDBOOT()			TRAP7(0)
#define WARMBOOT()			TRAP7(1)
#define GETCHAR()			TRAP7(2)
#define SGETCHAR()			TRAP7(3)
#define PUTCHAR(CH)			TRAP7(4, CH)
#define SPUTCHAR()			TRAP7(5)
#define GETC(DVR)			TRAP7(6, DVR)
#define SGETC(DVR)			TRAP7(7, DVR)
#define PUTC(DVR, CH)			TRAP7(8, DVR, CH)
#define SPUTC(DVR)			TRAP7(9, DVR)
#define ADD_IPDVR(IO, STAT, NAME, PV)	TRAP7(10, IO, STAT, NAME, PV)
#define ADD_XIPDVR(io, stat, name,pv,m)	TRAP7(10, io, stat, name, pv, m)
#define LOADREL(IFD, ADDR)		TRAP7(11, IFD, ADDR)
#define ADD_OPDVR(IO, STAT, NAME, PV)	TRAP7(12, IO, STAT, NAME, PV)
#define EXIT(I)				TRAP7(13, I)
#define SET_SIP(DVR)			TRAP7(14, DVR)
#define SET_SOP(DVR)			TRAP7(15, DVR)
#define SET_VSVEC(VEC, RATE, CALLVAL)	TRAP7(16, VEC, RATE, CALLVAL)
#define CLR_VSVEC(VNUM)			TRAP7(17, VNUM)
#define GET_TICKS()			TRAP7(18)
#define GET_CPU()			TRAP7(19)
#define SET_SER(DVR)			TRAP7(20, DVR)
#define CASWRAW(START, LENGTH, LEADER)	TRAP7(21, START, LENGTH, LEADER)
#define CASRRAW(BUF, LEADER, MAXHUNK)	TRAP7(22, BUF, LEADER, MAXHUNK)
#define GETDATE(STR)			TRAP7(23, STR)
#define SETDATE(STR)			TRAP7(24, STR)
#define ABORTSTAT()			TRAP7(25)
#define ENT1INTS(VEC, PRELOAD)		TRAP7(26, VEC, PRELOAD)
#define DIST1INTS()			TRAP7(27)
#define SINE(I)				TRAP7(28, I)
#define DEF_FK(FKNUM, STR)		TRAP7(29, FKNUM, STR)
#define GETRAND()			TRAP7(30)
#define SET_640(MODE)			TRAP7(31, MODE)
#define SET_VDP(PAGE)			TRAP7(32, PAGE)
#define SET_VAP(PAGE)			TRAP7(33, PAGE)
#define SET_FGCOL(COL)			TRAP7(34, COL)
#define SET_BGCOL(COL)			TRAP7(35, COL)
#define SET_BDCOL(COL)			TRAP7(36, COL)
#define SET_PAL(PALPOS, COL)		TRAP7(37, PALPOS, COL)
#define RDCH_SHAPE(CHNO)		TRAP7(38, CHNO)
#define DEF_CHSHAPE(CHNO, DEFPTR)	TRAP7(39, CHNO, DEFPTR)
#define DEF_WIND(WIND)			TRAP7(40, WIND)
#define VID_ADDRESS(X, Y)		TRAP7(41, X, Y)
#define MOVE_WIND(BUF, MODE)		TRAP7(42, BUF, MODE)
#define RAWVID(ROW, COL, CH, FG, BG)	TRAP7(43, ROW, COL, CH, FG, BG)
#define FILL_WIND(COL)			TRAP7(44, COL)
#define SCURS_MODE(RATE, EN, MASK)	TRAP7(45, RATE, EN, MASK)
#define MOUSETRAP(N, V)			TRAP7(46, N, V)

/* Extra mousetrap modes */
#define SAVEVCONTEXT(p)			TRAP7(46, 4, p)
#define RESTVCONTEXT(p)			TRAP7(46, 5, p)
#define VCONTEXTSIZE()			TRAP7(46, 6)

#define FILL(X, Y, VAL)			TRAP7(47, X, Y, VAL)
#define RSET_PEL(X, Y, VAL)		TRAP7(50, X, Y, VAL)
#define SET_PEL(X, Y, VAL)		TRAP7(51, X, Y, VAL)
#define RLINE(X0, Y0, X1, Y1)		TRAP7(52, X0, Y0, X1, Y1)
#define DRAWLINE(X0, Y0, X1, Y1)	TRAP7(53, X0, Y0, X1, Y1)
#define RREAD_PEL(X, Y)			TRAP7(54, X, Y)
#define READ_PEL(X, Y)			TRAP7(55, X, Y)
#define SGFGCOL(VAL)			TRAP7(56, VAL)
#define RCIRCLE(X, Y, RADIUS)		TRAP7(59, X, Y, RADIUS)
#define CIRCLE(X, Y, RADIUS)		TRAP7(60, X, Y, RADIUS)
#define SDOTMODE(MODE)			TRAP7(61, MODE)
#define SDOTMODE2(MODE, ARG)		TRAP7(61, MODE, ARG)
#define GETMEM(NBYTES, MODE)		TRAP7(62, NBYTES, MODE)
#define GETFMEM(FIRST, NBYTES, MODE)	TRAP7(63, FIRST, NBYTES, MODE)
#define FREEMEM(ADDR)			TRAP7(64, ADDR)
#define CHDIR(ADDR)			TRAP7(65, ADDR)
#define MKDIR(ADDR, N)			TRAP7(66, ADDR, N)
#define GETFULLPATH(PATH, MODE)		TRAP7(67, PATH, MODE)
#define PATHCMP(P1, P2)			TRAP7(68, P1, P2)
#define FLOADREL(PATH, GM)		TRAP7(69, PATH, GM)
#define ANIPSEL(IPNUM)			TRAP7(70, IPNUM)
#define ANOPSEL(OPNUM)			TRAP7(71, OPNUM)
#define ANOPDIS(V)			TRAP7(72, V)
#define ADC()				TRAP7(73)
#define DAC(VAL)			TRAP7(74, VAL)
#define SET_LED(VAL)			TRAP7(75, VAL)
#define FREETONE(TABLE, TABLEN, L, P)	TRAP7(76, TABLE, TABLEN, L, P)
#define FTTIME()			TRAP7(77)
#define	RDIPORT()			TRAP7(78)
#define RDBIPORT()			TRAP7(79)
#define SETSTVEC(VECNUM, WHERETO)	TRAP7(80, VECNUM, WHERETO)
#define NEW_CBUF(BNUM, ADDR, LEN)	TRAP7(81, BNUM, ADDR, LEN)
#define PROG_SIO(CHAN, SPPTR)		TRAP7(82, CHAN, SPPTR)
#define GETTDSTR(BUF)			TRAP7(83, BUF)
/* Extra (V4.2) modes for gettdstr */
#define CVTTDSTR(dateptr, mybuf)	TRAP7(83, 0, 0, dateptr, mybuf)
#define READTIMEINC()			TRAP7(83, 0, 1, 0)
#define SETTIMEINC(inc)			TRAP7(83, 0, 1, inc)

#define NLEDIT(STR, LEN)		TRAP7(84, STR, LEN)
#define BADSYSCALL()			TRAP7(85)
#define LEDIT(STR)			TRAP7(86, STR)
#define IEXEC(PROMPT)			TRAP7(87, PROMPT)
#define EXEC(LB)			TRAP7(88, LB)
#define CALLMRD(N, CMD, ARG)		TRAP7(89, N, CMD, ARG)
#define SET_KVEC(VEC)			TRAP7(90, VEC)
#define CLPARSE(A, T, V)		TRAP7(91, A, T, V)
#define QSORT(B, N, W, C)		TRAP7(92, B, N, W, C)
#define SLICEARGS(S, A, C)		TRAP7(93, S, A, C)
#define CPUSPEED()			TRAP7(94)
#define FIND_DRIVER(IORO, NAME)		TRAP7(95, IORO, NAME)
#define GET_DVRLIST(IORO)		TRAP7(96, IORO)
#define EXECA(ARGV)			TRAP7(97, ARGV)
#define EXECV(PATH, ARGV)		TRAP7(98, PATH, ARGV)
#define OPTION(OPNUM, SET)		TRAP7(99, OPNUM, SET)
#define INST_BDVR(BR, BW, BS, NAME, B)	TRAP7(100, BR, BW, BS, NAME, B)
#define FIND_BDVR(NAME)			TRAP7(102, NAME)
#define BLKREAD(BLK, BUF, DEV)		TRAP7(103, BLK, BUF, DEV)
#define BLKWRITE(BLK, BUF, DEV)		TRAP7(104, BLK, BUF, DEV)
#define OPEN(NAME, MODE)		TRAP7(105, NAME, MODE)
#define READ(FD, BUF, NBYTES)		TRAP7(106, FD, BUF, NBYTES)
#define CLOSE(FD)			TRAP7(107, FD)
#define CREAT(NAME, BITS, ADDR)		TRAP7(108, NAME, BITS, ADDR)
#define WRITE(FD, BUF, NBYTES)		TRAP7(109, FD, BUF, NBYTES)
#define UNLINK(NAME)			TRAP7(110, NAME)
#define RENAME(OLD, NEW)		TRAP7(111, OLD, NEW)
#define FILESTAT(NAME, BUF)		TRAP7(112, NAME, BUF)
#define READDIR(DEV, BUF, DP, POS, PAR)	TRAP7(113, DEV, BUF, DP, POS, PAR)
#define INTERPBEC(EC, BUF)		TRAP7(114, EC, BUF)
#define SEEK(FD, OFFSET, MODE)		TRAP7(115, FD, OFFSET, MODE)
#define TELL(FD)			TRAP7(116, FD)
#define BDMISC(BDNUM, CODE, ARG1)	TRAP7(117, BDNUM, CODE, ARG1)
#define PROCESSDIR(PN, BUF, MODE)	TRAP7(118, PN, BUF, MODE)
#define MULTIBLKIO(dvr,cmd,ptr,blk,nblks) TRAP7(119, dvr,cmd,ptr,blk,nblks)
#define FPUTS(FD, BUF)			TRAP7(121, FD, BUF)
#define ERRMES(EC)			TRAP7(122, EC)
#define FGETS(FD, BUF)			TRAP7(123, FD, BUF)
#define RDALLDIR(p, mm, sm, ps)		TRAP7(124, p, mm, sm, ps)
#define OSCONTROL(a1, a2)		TRAP7(125, a1, a2)
#define OSC1(a1)			TRAP7(125, a1)
#define GETROMVER()			TRAP7(126)
#define	LOCKZ80(mode, arg)		TRAP7(127, mode, arg)
#define AEXECA(argv, isasync)		TRAP7(128, argv, isasync)
#define PROCCNTL(mode, arg)		TRAP7(129, mode, arg)
#define NEWCHSET(PTR)			TRAP7(130, PTR)
#define SCHEDPREP(addr, av, flags, ss)	TRAP7(131, addr, av, flags, ss)
#define PIPE(ptr)			TRAP7(132, ptr)
#define CDMISC(dvr, cmd, p1, p2, p3)	TRAP7(133, dvr, cmd, p1, p2, p3)
#define SLINK(one, two, three)		TRAP7(134, one, two, three)
#define ALIAS(cmd, arg)			TRAP7(135, cmd, arg)
#define SPR(fmt, putcf, arg)		TRAP7(139, fmt, putcf, arg)
#define CRTC_INIT(mode, p)		TRAP7(140, mode, p)
#define CHKPERM(pdirent, mask, path)	TRAP7(141, pdirent, mask, path)
#define FNLEDIT(buf, len, in, out)	TRAP7(142, buf, len, in, out)

/* Various modes for proccntl */
#define GETPID()			TRAP7(129, 0)
#define GETPPID(PID)			TRAP7(129, 1, PID)
#define KILL(PID)			TRAP7(129, 3, PID)
#define SLEEP(TICKS)			TRAP7(129, 4, TICKS)
#define GETPROCTAB(PID)			TRAP7(129, 5, PID)
#define CWD(PID, PATH)			TRAP7(129, 6, PID, PATH)
#define LOCKIN(MODE)			TRAP7(129, 7, MODE)
#define RUNSTATS(mode)			TRAP7(129, 8, mode)
#define WAIT(PID)			TRAP7(129, 9, PID)
#define NICE(PID, V)			TRAP7(129, 10, PID, V)
#define SIGSEND(PID, V1, V2)		TRAP7(129, 11, PID, V1, V2)
#define SIGCATCH(V)			TRAP7(129, 12, V)
#define SENDINTERRUPT(RP)		TRAP7(129, 13, RP)
#define PROCTRACE(RP)			TRAP7(129, 14, RP)
#define GETPROCFLAGS(PID)		TRAP7(129, 15, PID)
#define ISINTERACTIVE(PID)		TRAP7(129, 16, PID)
#define NOSPCHECK(pid, mode)		TRAP7(129, 17, pid, mode)
#define CSVEC(v)			TRAP7(129, 18, v)
#define GETPCURPID()			TRAP7(129, 19)
#define READSIGVEC(pid)			TRAP7(129, 20, pid)
#define FSBPTR()			TRAP7(129, 21)
#define FSPPTR()			TRAP7(129, 22)
#define SSPTR()				TRAP7(129, 23)
#define	KILLUSER(hs)			TRAP7(129, 24, hs)
#define SIGBLOCK(pid, blk)		TRAP7(129, 25, pid, blk)
#define ALARM(n)			TRAP7(129, 26, n)
#define SIGBLOCKER(rootpid, sig, arg)	TRAP7(129, 27, rootpid, sig, arg)
#define SNOOZE(vec, p1, p2)		TRAP7(129, 28, vec, p1, p2)
#define SIGUSER(pid, sig, arg)		TRAP7(129, 29, pid, sig, arg)
#define FINDHOMESHELL(pid)		TRAP7(129, 30, pid)
#define SETSHNICE(pid, nice)		TRAP7(129, 31, pid, nice)
#define LASTCHILD(pid)			TRAP7(129, 32, pid)
#define SWPPTR()			TRAP7(129, 33)
#define KILLDOWN(STARTPID)		TRAP7(129, 34, STARTPID)
#define SIGDOWN(STARTPID, VAL1, VAL2)	TRAP7(129, 35, STARTPID, VAL1, VAL2)
#define KILLUID(uid)			TRAP7(129, 36, uid)
#define SIGUID(uid, p1, p2)		TRAP7(129, 37, uid, p1, p2)
#define SETSIGEXIT(pid, mode)		TRAP7(129, 38, pid, mode)
#define SETPG(pg)			TRAP7(129, 39, pg)
#define SIGPG(pg, sig, arg)		TRAP7(129, 40, pg, sig, arg)
#define KILLPG(pg)			TRAP7(129, 41, pg)
#define SETPROCUMASK(pid, mask)		TRAP7(129, 42, pid, mask)
#define SETENVBITS(pid, mask, set)	TRAP7(129, 43, pid, mask, set)
#define GETENVBITS(pid)			TRAP7(129, 44, pid)
#define NAMETOPID(pid)			TRAP7(129, 45, pid)
#define BLOCKTX(pid, addr, len, sig)	TRAP7(129, 46, pid, addr, len, sig)
#define BLOCKRX(mode)			TRAP7(129, 47, mode)
#define SETENV(pid, name, set, mode)	TRAP7(129, 48, pid, name, set, mode)
#define GETENV(pid, name, mode)		TRAP7(129, 49, pid, name, mode)
#define BGNICE(PID, V)			TRAP7(129, 50, PID, V)

/* The modes for oscontrol */
#define SETIBCVEC(vec)			OSCONTROL(0, vec)
#define OBRAMSTART()			OSC1(1)
/* No 2! */
#define READIBCVEC()			OSC1(3)
#define FORCELEVEL0()			OSC1(4)
#define READXPATH(n)			OSCONTROL(5, n)
#define READASSIGN(m)			OSCONTROL(6, n)
#define SETUMASK(n)			OSCONTROL(7, n)
#define READUMASK()			OSC1(8)
#define SETUID(n)			OSCONTROL(9, n)
#define READUID()			OSC1(10)
#define GET_BDEV(path)			OSCONTROL(11, path)
#define DUMPLASTLINES(cdev)		OSCONTROL(12, cdev)
#define SETWILDVEC(vec)			OSCONTROL(13, vec)
#define READWILDVEC()			OSC1(14)
#define VIDEO_INIT(level)		OSCONTROL(17, level)
#define KB_INIT(level)			OSCONTROL(18, level)
#define SETBEEPVOL(vol)			OSCONTROL(19, vol)
#define READBEEPVOL()			OSC1(20)
#define SETBEEPVEC(vec)			OSCONTROL(21, vec)
#define READBEEPVEC()			OSC1(22)
#define NOUSEFFBS(n)			OSCONTROL(23, n)
#define READFFBS()			OSC1(24)
#define SETMEMFAULT(n)			OSCONTROL(25, n)
#define RXTXPTR(n)			OSCONTROL(26, n)
#define SETBDLOCKIN(n)			OSCONTROL(27, n)
#define READBDLOCKIN()			OSC1(28)
#define STARTOFDAY()			OSC1(29)
#define PFASTCHAR()			OSC1(30)
#define SRBRCLOCK(n)			OSCONTROL(31, n)
#define TIMER1USED()			OSC1(32)
#define TRASHASSIGN(uid)		OSCONTROL(33, uid)
#define TRASHENVSTRINGS(pid)		OSCONTROL(34, pid)
#define ENVSUB(s, dollaronly, mmode)	TRAP7(125, 35, s, dollaronly, mmode)
#define DOASSIGN(argc, argv)		TRAP7(125, 36, argc, argv)
#define PSYSTICKS()			TRAP7(125, 37)
#define OBRAMSIZE()			TRAP7(125, 38)

/*
 * Read entry points from ROM to use more arguments
 */

#define PRINTF		(*(int(*)())SETSTVEC(48, -1))
#define SPRINTF		(*(int(*)())SETSTVEC(49, -1))
#define FPRINTF		(*(int(*)())SETSTVEC(120, -1))
