/*
 * Serial I/O drivers
 */

#include <types.h>
#include <syscalls.h>
#include <hwdefs.h>
#include <storedef.h>
#include <scc.h>
#include <kt/duart.h>
#include <kt/gpiomap.h>
#include "vectors.h"
#include <ss/simvec.h>
#include <process.h>
#include <splfns.h>

#define CONTROLO ('O' - 64)

ushort abort_flag;		/* True if abort hit */
int stoplist;
ushort contohit;
uchar rxamask, rxbmask;

int (*disr)();
uchar intregs2;

/* Default programming for serial ports */
/* baud, rxbits, txbits, parity, stop */
struct scc_prog def_asprog = { 9600, 3, 3, 0, 1 };
struct scc_prog def_bsprog = { 9600, 2, 2, 2, 2 };
struct scc_prog def_fbsprog = { 19200, 2, 2, 2, 2 };
struct scc_prog curprogs2;	/* Current settings */

struct circ_buf da_icbuf, db_icbuf;
struct circ_buf da_ocbuf, db_ocbuf;
char da_ibufSER_BUFSIZE;
char db_ibufSER_BUFSIZE;
char da_obufSER_BUFSIZE;
char db_obufSER_BUFSIZE;

comms_init(reslevel)
{
	int chanbrx();
	register int chan;
	struct scc_prog *psccp;

	intregs0 = intregs1 = 0;
	stoplist = 0;
	NEW_CBUF(0, 0, 0);	/* Set up default circular buffers */
	NEW_CBUF(1, 0, 0);
	NEW_CBUF(2, 0, 0);
	NEW_CBUF(3, 0, 0);
	contohit = 0;
	*(int *)CHANBVEC = (int)chanbrx;
	for (chan = 0; chan < 2; chan++)
	{
		if (chan == 0)
			psccp = &def_asprog;
		else
			psccp = (*IPORT & 1) ? &def_bsprog : &def_fbsprog;
		if ((reslevel == 0) || (prog_sio(chan, &curprogschan) < 0))
			PROG_SIO(chan, psccp);
	}
}

/*
 * System call 82: Program a serial port. Return non-zero if bad data supplied
 */

_prog_sio(chan, spptr)
uint chan;
register struct scc_prog *spptr;	/* All the data */
{
	int retval;
	int  du_aisr();			/* Place to vector to! */

	if (chan > 1)
		return -1;
	if (spptr == 0)
		return (int)&curprogschan;

	retval = duart_init(0, chan, spptr->brate, spptr->rxbits + 5,
			spptr->parity, D_INTVEC, 1, 0, du_aisr);
	if (retval >= 0)
		movmem(spptr, &curprogschan, sizeof(*spptr));
	return retval;
}

/* Return input status */
da_istat() { return do_istat(&da_icbuf); }
db_istat() { return do_istat(&db_icbuf); }

/* Return a char from channel */
da_getc() { return get_cbuf(&da_icbuf); }
db_getc() { return get_cbuf(&db_icbuf); }

/* Return output status */
da_ostat() { return DUART->d_achnl.d_status & TXRDY; }
db_ostat() { return DUART->d_bchnl.d_status & TXRDY; }

/* Send a char to channel */
da_putc(ch)
{
	while (!(DUART->d_achnl.d_status & TXRDY))
		;
	DUART->d_achnl.d_chbuf = ch;
}

db_putc(ch)
{
	while (!(DUART->d_bchnl.d_status & TXRDY))
		;
	DUART->d_bchnl.d_chbuf = ch;
}

/*
 * Return status of an input circular buffer (number of buffered chars)
 */

do_istat(cbptr)
register struct circ_buf *cbptr;
{
	register int s, nch;

/*	s = DUART_SPL();*/
	nch = cb_bnum(cbptr);
/*	splx(s);*/
	return nch;
}

/*
 * Transmit a char to a channel. If the Tx buffer is empty, force transmission
 * of a char from the channel
 */

#ifdef notdef
do_putc(dptr, cbptr, ch)
register struct d_chnl *dptr;			/* The I/O port */
register struct circ_buf *cbptr;
char ch;
{
	int s;

	wroom_cbuf(cbptr);		/* Wait until there is a gap */
	s = DUART_SPL();
	put_cbuf(cbptr, ch);
	if (dptr->d_status & TXRDY)	/* Buffer empty? */
		do_tx(dptr, cbptr);
	splx(s);
	return 0;
}

/*
 * General Transmit ISR code. If there are chars in buffer, send one, else
 * reset pending Tx interrupt.
 */

do_tx(dptr, cbptr)
register struct d_chnl *dptr;			/* The channel */
register struct circ_buf *cbptr;	/* The circular buffer structure */
{
	if (cbptr->head != cbptr->tail)	/* Any chars there? */
	{
		dptr->d_chbuf = get_cbuf(cbptr);
	}
	else				/* Reset Tx int pending command */
		dptr->d_cmd = M_NOP|RX_NOP|TX_STOP;
}
#endif

/*
 * System call 25: Return abort flag status
 */

_abortstat()
{
	ushort x;

	x = abort_flag;
	abort_flag = 0;
	return x;
}

du_isr()
{
	while (DUART->d_isr & BRX_RDYFF)
		(*(int (**)())CHANBVEC)(DUART->d_bchnl.d_chbuf & rxbmask);
	while (DUART->d_isr & ARX_RDYFF)
	{	/* Buffer the char */
		put_cbuf(&da_icbuf, DUART->d_achnl.d_chbuf & rxamask);
	}
}

chanbrx(ch)
{
	register int *rptr;
	register int i;

	if (contohit)
	{	/* Last char was a ^O */
		stoplist = 0;
		contohit = 0;
		if (ch < 32)	/* Control char? */
			ch += 64;
		switch (toupper(ch))
		{
		case 'R':	/* Reset */
			WARMBOOT();
		case 'C':	/* Abort */
			abort_flag = 1;
			return;
		case 'D':	/* Register dump */
			comms_init();
			rptr = &ch;
			PRINTF("\r\nD0-D7:");
			for (i = 0; i < 8; i++)
				PRINTF(" %08x", *rptr++);
			PRINTF("\r\nA0-A7:");
			for (i = 0; i < 7; i++)
				PRINTF(" %08x", *rptr++);
			PRINTF(" %08x\r\n", &ch);
			PRINTF("SR: %04x  ", *(ushort *)rptr);
			rptr = (int *)((int)rptr + 2);
			PRINTF("PC: %08x\r\n", *rptr);
			return;
		default:
			return;
		}
	}
	else if (ch == CONTROLO)
	{
		stoplist = 1;
		contohit = 1;
	}
	else
		put_cbuf(&db_icbuf, ch);
}

duart_init(dnum, chan, brate, nbits, parity, vector, rxinten, txinten, isr)
int dnum, chan, brate, nbits, parity, vector, rxinten, txinten;
int (*isr)();
{
	register struct duart *dptr;
	register struct d_chnl *dchp;
	int s;
	int pa, nb, br, bitmask;

	disr = isr;			/* Where to go to */

	switch (brate)
	{
	case 75: br = S2B75; break;
	case 110: br = S2B110; break;
	case 150: br = S2B150; break;
	case 300: br = S2B300; break;
	case 600: br = S2B600; break;
	case 1200: br = S2B1200; break;
	case 2000: br = S2B2000; break;
	case 2400: br = S2B2400; break;
	case 4800: br = S2B4800; break;
	case 1800: br = S2B1800; break;
	case 9600: br = S2B9600; break;
	case 19200: br = S2B19200; break;
	default: return -1;
	}
	switch (nbits)
	{
	case 5:	nb = BITS5; bitmask = 0x1f; break;
	case 6: nb = BITS6; bitmask = 0x3f; break;
	case 7: nb = BITS7; bitmask = 0x7f; break;
	case 8: nb = BITS8; bitmask = 0xff; break;
	default: return -1;
	}
	switch(parity)
	{
	case 0:	pa = PAROFF; break;
	case 1: pa = PARON|ODDPAR; break;
	case 2: pa = PARON|EVENPAR; break;
	default: return -1;
	}
	if (!chan)
		rxamask = bitmask;
	else
		rxbmask = bitmask;

	dptr = DUART;
	dchp = chan ? &dptr->d_bchnl : &dptr->d_achnl;
	bitmask = chan ? 0x0a : 0x05;

	s = spl7();

	if (vector)
	{
		int  du_aisr();			/* Place to vector to! */

		*((int *)(vector*4)) = (int)du_aisr;
		dptr->d_ivec = vector;			/* Interrupt vector */
	}
	dptr->d_opcr = 0;			/* Normal outputs */
	dptr->d_aux = BRG_SET2|TM_X16XTAL;	/* Timer mode  */

	dchp->d_cmd = M_TXRESET; 		/* Clean things up */
	dchp->d_cmd = M_RXRESET; 		/* Clean things up */
	dchp->d_cmd = M_GETMR1|TX_NOP|RX_NOP;
	dchp->d_mode = nb|pa|ERROR_CHAR|RX_RDY;
	dchp->d_mode = STOPB1|TXUSE_CTS|TXIGN_RTS|OPS_NORMAL;
	/*    Set up the Baud Rate Generator   */
	dchp->d_clock = (br << 4) | br;
	dchp->d_cmd = M_NOP|TX_GO|RX_GO;

	dptr->d_obset = bitmask;		/* Turn on RTS/DTR */
	dptr->d_aux = BRG_SET2|TM_X16XTAL;

	/* Enable interrupts */
	if (rxinten)
		intregsdnum |= chan ? BRX_RDYFF : ARX_RDYFF;
	else
		intregsdnum &= ~(chan ? BRX_RDYFF : ARX_RDYFF);
	if (txinten)
		intregsdnum |= chan ? BTX_RDY : ATX_RDY;
	else
		intregsdnum &= ~(chan ? BTX_RDY : ATX_RDY);
	DUART->d_intrg = intregs0;
	splx(s);
	return 0;
}
