/* TABS4 NONDOC
 * Keyboard driver
 */

#include <types.h>
#include <syscalls.h>
#include <hwdefs.h>
#include <via.h>
#include <storedef.h>
#include <options.h>
#include <chario.h>
#include <splfns.h>

/* Define special ALT keys */
#define CTL(c)		((c)-0x40)

#define AKY_CASMOT	't'			/* Toggle cassette motor */
#define AKY_ABORT	'C'			/* Set the abort flag */
#define AKY_FWDARROW	0x7f	/* Delete key on keypad */

/* Bits for left & right shift key flags */
#define LEFTSHIFT	1
#define RIGHTSHIFT	2

/* Definitions for special keys (top bits set) */
#define K_CTRL	0x80		/* Control key */
#define K_LS	0x81		/* Left shift */
#define K_RS	0x82		/* Right shift */
#define K_PSC	0x83		/* Prt Sc */
#define K_ALT	0x84		/* Alternate */
#define K_CL	0x85		/* Caps lock */
#define K_NL	0x86		/* Num lock */
#define K_SL	0x87		/* Scroll lock */
#define K_BAD	0xff		/* An invalid key has been hit */
#define NNKPK	13			/* Number of numeric keypad keys */
#define NFKEYS	12			/* Number of function keys */
#define FKSIZE	64			/* Max size of function keys */

ushort abort_flag;

uchar fkeysNFKEYSFKSIZE;	/* The function key table */
uchar *capture;				/* The function key pointer for capture mode */
int capindex;				/* Index into above */
unsigned short capkey;		/* The FK currently being captured */
char rawcodes;				/* If true, scancodes sent directly */
struct circ_buf kb_cbuf;	/* The type-ahead buffer and structure */
int (*keyscan_vec)();		/* The vector to the key scan handler */

char kb_vi_mode;			/* Flag: set keyboard into VI mode */

/* The numeric keypad number which is calculated when ALT is held */
short spellascii;

extern char rootshellname;

uchar key_tab =
{
	0x1b,	'1',	'2',	'3',	'4',	'5',	'6',	/* 1-7 */
'7',	'8',	'9',	'0',	'-',	'=',	'\b',	'\t',	/* 8-15 */
'Q',	'W',	'E',	'R',	'T',	'Y',	'U',	'I',	/* 16-23 */
'O',	'P',	'',	'',	'\r',	K_CTRL,	'A',	'S',	/* 24-31 */
'D',	'F',	'G',	'H',	'J',	'K',	'L',	';',	/* 32-39 */
0x27,	'`',	K_LS,	0x5c,	'Z',	'X',	'C',	'V',	/* 40-47 */
'B',	'N',	'M',	',',	'.',	'/',	K_RS,	'*'/*K_PSC*/,	/* 48-55 */
K_ALT,	' ',	K_CL,	0xc0,	0xc1,	0xc2,	0xc3,	0xc4,	/* 56-63 */
0xc5,	0xc6,	0xc7,	0xc8,	0xc9,	K_NL,	K_SL,	0xd0,	/* 64-71 */
0xd1,	0xd2,	0xd3,	0xd4,	0xd5,	0xd6,	0xd7,	0xd8,	/* 72-79 */
0xd9,	0xda,	0xdb,	0xdc,	K_BAD,	K_BAD,	K_BAD,	0xca,	/* 80-87 */
0xcb								/* 88-88 */
};

#define NKEYS sizeof(key_tab)		/* The number of defined keys */

/* Map the codes which the keys in the range ' to = produce when shifted */
static uchar oddmap =
{
	'"',	K_BAD,	K_BAD,	K_BAD,	K_BAD,	'<',	'_',	'>',
	'?',	')',	'!',	'@',	'#',	'$',	'%',	'^',
	'&',	'*',	'(',	K_BAD,	':',	K_BAD,	'+'
};

/* The codes generated by the numeric keypad when numlock is active */
static uchar nkp_ltab =  {
	'7','8','9','-','4','5','6','+','1','2','3','0','.'};

/* The codes generated by the numeric keypad when numlock is inactive */
static uchar nkp_nltab =
	{
		CTL('L'),	CTL('E'),	CTL('R'),	'-',
		CTL('S'),	'5',		CTL('D'),	'+',
		CTL('B'),	CTL('X'),	CTL('C'),	CTL('\\'),	0x7f
	};

/* The codes generated by the numeric keypad when numlock is inactive  in VI mode */
static uchar VInkp_nltab =
	{
		'H',	'k',	CTL('U'),	'-',
		'h',	'5',	'l',	'+',
		'L',	'j',	CTL('D'),	'i',	'x'
	};

static revbyte();

EXTERN int curpid;
EXTERN int miscmodes;

/*
 * Return number of keys buffered for K/B
 */

kb_istat()
{
	return cb_bnum(&kb_cbuf);
}

/*
 * Get a key from the K/B
 */

kb_getc()
{
	return get_cbuf(&kb_cbuf);
}

/*
 * VIA shift register interrupt entry point: Get scan code & vector to code
 */

kb_cisr()
{
	int key;

	key = *V_SR;
	clrkblatch();
	if (key == 0x07 || key == 0x87)	/* Reversed 0xe0 & 0xe1 */
		return;

	entry(keyscan_vec, revbyte(key));
}

/* Clear U53 */
clrkblatch()
{
	*V_DDRA |= CLRKBL;	/* Ensure it is pointing the correct way */
	*V_AREG &= ~CLRKBL;
	*V_AREG |= CLRKBL;
}

/*
 * System call 90: Install the passed keyscan vector, return old one.
 * Set up default if input == 0.
 * If vec == 1, set rawcodes, return old value.
 * If vec == 2, clear it, return old value.
 * If vec == 3, read old value.
 */

_set_kvec(vec)
register int (*vec)();
{
	register int retval, s;
	register uint ivec;
	int key_proc();

	retval = rawcodes;
	if (vec == (int (*)())1)
	{
		CDMISC(0, CDM_SETRAWMODE, 1);
		rawcodes = 1;
	}
	else if (vec == (int (*)())2)
	{
		CDMISC(0, CDM_SETRAWMODE, 0);
		rawcodes = 0;
	}
	else if (vec == (int (*)())3)
		return retval;
	else
	{
		ivec = (unsigned int)vec;
		retval = (int)keyscan_vec;
		s = VIA_SPL();
		keyscan_vec = (vec) ? vec : key_proc;
		splx(s);
	}
	return retval;
}

/*
 * Default keyboard scan code interpreter
 */

key_proc(key)
register uchar key;
{
	register unsigned short keycode, retkey;
	register uchar *ptfk;

	if (rawcodes)
	{
		putkey(key);
		return key;
	}

/* Kludge here for the wierd codes coming from Kathy's Darwin K/B set */
	if ((key == 0xe0) || (key == 0xe1) || (key == 0x6f) || (key == 0xef))
		return;

	key--;

	if ((key & 0x7f) > NKEYS)	/* Bad key? */
	{
		retkey = K_BAD;
		goto gotkey;
	}
	keycode = key_tabkey & 0x7f;

	if (key & 0x80)		/* Is it a release? */
	{	/* Was a special key released? */
		if ((!(keycode & 0x80)) || (keycode > K_SL))
			return;
		switch (keycode)	/* A special key has been released */
		{
		case K_CTRL:
			ak_ctrl = 0;
			break;
		case K_LS:
			ak_shift &= ~LEFTSHIFT;
			break;
		case K_RS:
			ak_shift &= ~RIGHTSHIFT;
			break;
		case K_PSC:
			break;
		case K_ALT:
			if (spellascii >= 0)
				putkey(spellascii);
			ak_alt = 0;
			break;
		case K_CL:
			if (!ak_ctrl)
				ak_capslock = !ak_capslock;
			break;
		case K_NL:
			if (!ak_ctrl)
				ak_numlock = !ak_numlock;
			break;
		case K_SL:
			break;
		default:	/* Shouldn't happen! */
			KB_INIT(3);
			break;
		}
		return;
	}

/* A key has been depressed */
	if (keycode & 0x80)	/* Was it a special key? */
	{	/* Was it on the NKP? */
		if (keycode > (0xd0 + NNKPK))
			retkey = K_BAD;
		else if (keycode >= 0xd0)
		{
			ushort ch;
			/* Index into NKP tables */
			ch = nkp_ltabkeycode - 0xd0;
			if (ak_alt)
			{	/* Could be spelling ascii */
				if ((ch >= '0') && (ch <= '9'))
				{
					if (spellascii < 0)
						spellascii = 0;	/* Start off */
					spellascii = spellascii * 10 + ch - '0';
					return;
				}
			}
			if ((ak_numlock && !ak_shift) || (!ak_numlock && ak_shift))
				retkey = ch;
			else
			{
				retkey = (kb_vi_mode ? VInkp_nltab : nkp_nltab)keycode - 0xd0;
				if ((retkey == '5') && (miscmodes & OPM_NKPCQ))
					retkey = CTL('Q');
			}
		}
		else
		{	/* Was it a function key? */
			if ((keycode >= 0xc0) && (keycode < (0xc0 + NFKEYS)))
			{
				keycode -= 0xc0;	/* Make it FK number */
				if (capture && (keycode == capkey))
				{	/* Defining terminated */
					*capture = 0;
					capture = 0;
					return;
				}
				if (ak_alt && ak_ctrl)
				{	/* Prepare to program FK */
					if (capture)
						*capture = 0;	/* Terminate */
					capture = fkeyskeycode;
					capindex = 0;
					capkey = keycode;
					return;		/* Start defining */
				}
				/* Expand the function key */
				ptfk = fkeyskeycode;
				while (*ptfk)
				{
					putkey((*ptfk == 0x80) ? 0 : *ptfk);
					ptfk++;
				}
				return;
			}
			else
			{	/* It is a special key */
				switch (keycode)
				{
				case K_CTRL:
					ak_ctrl = keycode; /* Use a register to set */
					break;
				case K_LS:
					ak_shift |= LEFTSHIFT;
					break;
				case K_RS:
					ak_shift |= RIGHTSHIFT;
					break;
				case K_PSC:
					break;
				case K_ALT:
					ak_alt = keycode;
					spellascii = -1;
					break;
				case K_CL:
					break;
				case K_NL:
					break;
				case K_SL:	/* Scroll lock */
					putkey('s' + 0x80);
					break;
				default:
					KB_INIT(3);
					break;
				}
				return;		/* Set the flag & go */
			}
		}
	}
	else		/* It is not a special key */
	{
		retkey = keycode;
		if (keycode > 'Z')	/*  to  group */
		{
			if (ak_shift)
			{
				if (keycode == '`')
					retkey = '~';
				else
					retkey += 0x20;
			}
			else if (ak_ctrl)
				retkey -= 0x40;
		}
		else if (keycode >= '@')
		{
			if (ak_ctrl)	/* Control key */
				retkey -= 0x40;
			else
			{
				if ((!ak_capslock && !ak_shift) || (ak_capslock && ak_shift))
					retkey += 0x20;
			}
		}
		else if (keycode >= 0x27)	/* Odds and ends */
		{
			if (ak_shift)
				retkey = oddmapkeycode - 0x27;
			if (ak_ctrl)
			{
				if (retkey == '_' || retkey == '^' || retkey == '@')
					retkey -= 0x40;
			}
		}
	}

/* We have the key - return with bit 7 set if ALT is depressed */
gotkey:
	if (retkey == K_BAD)
		KB_INIT(3);		/* Impossible code: clean it up */
	else
	{
		if (ak_alt)	/* ALT is depressed: look for special functions */
		{
			switch (retkey)
			{
			case AKY_CASMOT:	/* Toggle cassette motor */
				key = CAS_SPL();
				*AMUXLATCH = (alval ^= CASMOTOR);
				splx(key);
				break;
			case AKY_ABORT:		/* User abort (ALT-C) */
				abort_flag = 1;
				break;
			case AKY_FWDARROW:	/* Toggle eof char */
				toggleeof();
				break;
			default:		/* Buffer alternate key code */
noalt:				putkey(retkey | 0x80);
				break;
			}
		}
		else
			putkey(retkey);
	}
}

/*
 * Put an ascii char in the keyboard buffer.
 * If we are capturing to a function key, put it in the definition
 */

putkey(ch)
uchar ch;
{
	put_cbuf(&kb_cbuf, ch, 0);
	if (capture && !rawcodes)
	{
		*capture++ = ch;
		capindex++;
		*capture = '\0';	/* Terminate the string */
		if (capindex >= (FKSIZE - 1))
			capture = 0;	/* Buffer full */
	}
}

/*
 * Initialisation code. If reset level == 3, don't rewrite scan vector, etc
 */

kb_init(reslevel)
{
	register int zero = 0;
	register uchar *vareg = V_AREG;
	int s, fk;

	s = VIA_SPL();
/* Set up the vectors */
	*((int *) VIV_SR) = (int) kb_cisr;	/* Raw K/B input */
	ak_ctrl = zero;
	ak_alt = zero;
	ak_shift = zero;
	ak_capslock = zero;
	ak_numlock = zero;
	ABORTSTAT();						/* Clear flags */

	if (reslevel < 3)
	{
		SET_KVEC(zero);					/* Scan code interpreter */
		kb_vi_mode = zero;
		rawcodes = zero;
		capture = (uchar *)zero;		/* Not capturing */
	/* Initialise the structure to defaults */
		NEW_CBUF(5, zero, zero);
	}

	if (!reslevel)
		for (fk = 0; fk < NFKEYS; fk++)	/* Set up array of FK pointers */
			DEF_FK(fk, "");

	*V_ACR = (*V_ACR & 0xe3) | 0x0c;	/* SR shift in under CB1 */
	*V_DDRA |= CLRKBL|KBRESET;		/* K/B clear O/P bit */
	*vareg |= CLRKBL|KBRESET;
	clrkblatch();
	*V_SR = zero;				/* Read SR to start up */

/* Reset the K/B until it sends something */

	fk = 5;
	do
	{
		*vareg &= (uchar)~KBRESET;	/* Drop K/B reset */
		delay(300);					/* Delay 30 msec for K/B reset */
		*vareg |= (uchar)KBRESET;
		delay(300);					/* Delay for reset K/B */
	}
	while (!(*V_IFR & 4) && fk--);
#ifdef notdef
	if (fk < 1)
		FPRINTF(0, "K/B failure\r\n");
#endif
	*vareg |= KBRESET;			/* Unreset K/B */
	*V_DDRA &= ~KBRESET;			/* Set bit to I/P */

	*V_SR = zero;				/* Clear the pending interrupt */
	clrkblatch();
	*V_IER = 0x84;				/* Enable SR interrupts */

	splx(s);
}

/*
 * System call 25: Return abort flag status.
 * Replaced by signals in V4.2.
 */

_abortstat()
{
	ushort ret;

	ret = abort_flag;
	abort_flag = 0;
	return ret;
}

/*
 * System call 29: Define a function key
 */

_def_fk(fknum, fkaddr)
uint fknum;
uchar *fkaddr;
{
	if (fknum < NFKEYS)
	{
		movstr(fkeysfknum, fkaddr, FKSIZE);
		return 0;
	}
	else
		fknum -= 64;
	return (int)fkeysfknum;
}

/* Reverse the bits in a byte */

static
revbyte(num)
register unsigned short num;
{
	register unsigned short mask1, mask2, retval;

	retval = 0;
	mask1 = 0x80;
	mask2 = 0x01;
	do
	{
		if (num & mask1)
			retval |= mask2;
		mask1 >>= 1;
		mask2 <<= 1;
	}
	while (mask1);
	return retval;
}
