; TABS8
		psect	text
;
; Applix 1616/OS ROM assembler startup code
;
; Copyright (C) 1987 by Applix pty limited
;
; 28 Jan 1987. Programmer: Andrew K.P. Morton
;

BERR_VEC	equ	8		; Bus error vector
ILLEGALVEC	equ	16		; Illegal instruction vector

ROMSTART	equ	$500000	; Start of ROMs

CENTLATCH	equ	$600001	; Centronics latch
DACLATCH	equ	$600081	; DAC latch
VIDLATCH	equ	$600101	; Video latch
AMUXLATCH	equ	$600181	; Analog multiplexor latch
PAL0		equ	$600000	; Pallette entries
PAL1		equ	$600020
PAL2		equ	$600040
PAL3		equ	$600060

SCCBASE		equ	$700000	; SCC start
IPORT		equ	$700081	; Input port
VIABASE		equ	$700100	; VIA start
CRTCBASE	equ	$700180	; CRTC start

SCCBC		equ	SCCBASE+0	; SCC B control
SCCBD		equ	SCCBASE+2	; SCC B data
SCCAC		equ	SCCBASE+4	; SCC A control
SCCAD		equ	SCCBASE+6	; SCC A data

CRTCADDR	equ	CRTCBASE	; CRTC address register
CRTCDATA	equ	CRTCBASE+2	; CRTC data register

V_BREG		equ	VIABASE		; Port B I/O register	*/
V_AREG		equ	VIABASE+2	; Port A I/O register	*/
V_DDRB		equ	VIABASE+4	; Port B DDR		*/
V_DDRA		equ	VIABASE+6	; Port A DDR		*/

;
; Start of code
;

		global	_initsp,_main,__coldboot,__warmboot,_obramstart
		global	_pwruplongs,_asbootup,_excepsyscall
		global	_lastsyscall,_obramsize
		global	_user_patch

/* #define MDEBUG */

#ifdef MDEBUG
	global	_Tloc
	.comm	_Tloc,4

#define D(nn)				\
	move.l	a0,$4100	!	\
	move.l	_Tloc,a0	!	\
	move.l	#nn,(a0)+	!	\
	move.l	a0,_Tloc	!	\
	move.l	$4100,a0
#else
#define D(nn)
#endif

;__warmboot:
_asbootup:
bootup:		bra	tootle
_user_patch:	dc.l	0		; This is for user-patched ROMs
tootle:
#ifdef MDEBUG
	move.l	#$11000,_Tloc
#endif
		move.b	d0,ROMSTART	; Bank out ROMs
		or.w	#$0700,sr	; spl7
		move.l	_initsp,a7	; Set up Stack pointer (non-reset entry)
		reset
D(1)
dofn:
		move.b	IPORT,d0	; If switch 3 open
		bmi	toc		; set things up and branch to C

		move.b	d0,DACLATCH	; Echo it
		lsr.b	#4,d0		; Ignore bottom 4 bits
		and.l	#7,d0		; Only use 3 bits
		asl.l	#2,d0
		move.l	#ijtab,a0
		move.l	0(a0,d0.w),a0	; Get code address
		jmp	(a0)		; Go to it

__coldboot:	clr.l	_pwruplongs	; Ensure level 0 reset: zap the power up words
		jmp	__warmboot

ijtab:		dc.l	latchtest	; 0: Test analog, video, dac, cent, pallette
		dc.l	sccrwtest	; 1: SCC
		dc.l	viarwtest	; 2: VIA
		dc.l	crtcrwtest	; 3: CRTC
		dc.l	iporttest	; 4: IPORT
		dc.l	memrwtest	; 5: Read / write memory
		dc.l	memtest		; 6: Memory test
		dc.l	vidtest		; 7: Video test

; Write incrementing pattern to latches, pallette

latchtest:	clr.b	d3		; LED / relay bits
lt3:		move.l	#30000,d1	; Loop counter

lt2:		move.b	d0,CENTLATCH
		move.b	d0,DACLATCH
		move.b	d0,VIDLATCH

		move.b	d0,d2
		and.b	#$77,d2	; Preserve LED & relay bits
		or.b	d3,d2
		move.b	d2,AMUXLATCH

		move.b	d0,PAL0
		move.b	d0,PAL1
		move.b	d0,PAL2
		move.b	d0,PAL3
		addq.b	#1,d0
		subq.l	#1,d1
		bne	lt2
		eor.b	#$88,d3
		bra	lt3

; Read/write SCC

sccrwtest:	move.l	#SCCBASE,a0
		bra	rwtest

; Read/write VIA

viarwtest:
		move.b	#$fe,V_DDRA		; Port A has one input bit.
		move.b	#$ff,V_DDRB
vrwtest:	clr.b	d2
vrwtest3:	move.l	#100000,d1
vrwtest2:	move.b	d0,V_AREG
		move.b	d0,V_BREG
		move.b	V_BREG,DACLATCH		; Echo the read back byte
		addq.b	#1,d0
		subq.l	#1,d1
		bne.b	vrwtest2
		eor.b	#$8,d2
		move.b	d2,AMUXLATCH
		bra.b	vrwtest3

; Read/write CRTC

crtcrwtest:	move.l	#CRTCBASE,a0
		bra	rwtest

; Read from Input Port

iporttest:	move.l	#IPORT,a0
		bra	rtest

; Read / write memory

memrwtest:	move.l	#$55aa,a0
		bra	rwtest

; General input port test: read from port and write to Centronics latch.
; Input port address in A0.

rtest:		clr.b	d1
rtest3:		move.l	#100000,d0
rtest2:		move.b	(a0),DACLATCH
		subq.l	#1,d0
		bne.b	rtest2
		eor.b	#$8,d1
		move.b	d1,AMUXLATCH		; Flash LED
		bra.b	rtest3

; General output port test: increment latch contents. Latch address in A0

wtest:		clr.b	d2
wtest3:		move.l	#100000,d1
wtest2:		move.b	d0,(a0)
		addq.b	#1,d0
		subq.l	#1,d1
		bne.b	wtest2
		eor.b	#$8,d2
		move.b	d2,AMUXLATCH
		bra.b	wtest3

;General read/write test. Write a value to port, read it back, echo to centronics port,
;increment value & loop

rwtest:		clr.b	d2
rwtest3:	move.l	#100000,d1
rwtest2:	move.b	d0,(a0)
		move.b	(a0),DACLATCH
		addq.b	#1,d0
		subq.l	#1,d1
		bne.b	rwtest2
		eor.b	#$8,d2
		move.b	d2,AMUXLATCH
		bra.b	rwtest3

; Do memory test (this simplistic test will shortly be replaced)

memtest:	clr.w	d0
		clr.w	d1
		clr.w	d2		; Start value

memtest2:	move.w	d2,d0
		move.l	#0,a0
fillmem:	move.w	d0,(a0)+
		addq.w	#1,d0
		cmp.l	#$80000,a0
		bne.b	fillmem

		move.w	d2,d0
		move.l	#0,a0
checkmem:	cmp.w	(a0)+,d0
		bne.b	badmem
		addq.w	#1,d0
		cmp.l	#$80000,a0
		bne.b	checkmem

		add.w	#$101,d2	; New pattern	
		eor.b	#$88,d1
		move.b	d1,AMUXLATCH
		bra	memtest2

badmem:		subq.l	#2,a0		; Point to bad word
		move.l	a0,d0
		move.b	d0,DACLATCH	; Low byte of address
		lsr.l	#8,d0
		move.b	d0,VIDLATCH	; Next byte of address
		lsr.l	#8,d0
readbad2:	move.b	d0,AMUXLATCH	; Bits 16-23
		move.l	#30000,d1
readbad:	move.w	(a0),d2
		subq.l	#1,d1
		bne.b	readbad
		eor.b	#8,d0
		bra.b	readbad2

;
; Do Video test
;
		global	_cvidtest
vidtest:
		moveq.l	#1,d7
		bra.b	toc2

;vidtest:	move.l	#vtnext,a5
;		bra	crtcstart
;vtnext:
;		move.l	#$40000,a7	; Middle of RAM
;		bsr	getcpu		;Ascertain CPU type
;		bsr	setupram
;		clr.l	_obramst:w
;		jmp	_cvidtest

;
; Code executed when switch 3 not in test mode
;

toc:		clr.l	d7
toc2:		move.l	#tocnext,a5
 D(2)
		bra	crtcstart
tocnext:
 D(3)
		bsr	getcpu		;Ascertain CPU type
 D(4)
		clr.l	d3		; Search for start of on-board RAM
btloop:		add.l	#$00100000,d3	; Go in 1 Meg steps
		move.l	 d3,-(a7)
 D(5)
		bsr	_berrtest	; See if there is RAM there
 D(6)
		addq.l	#4,a7		; Adjust stack
		tst.w	d0		; Returns 1 if bus error
		bne.b	noram
		cmp.l	#$00500000,d3	; Do we have a full complement?
		bne.b	btloop
noram:		sub.l	#$00100000,d3	; Start of on-board RAM
		move.l	d3,_obramstart
 D(7)
; New stuff for 1/2 meg expansion
		move.l	#$80000,d0
		move.l	d0,_obramsize	; Assume 1/2 meg only.
		move.l	d3,a0		; Point to start of on-board RAM
		move.l	(a0),a2		; Save what's there
		clr.l	(a0)
		move.l	0(a0,d0.l),a1	; Save what's there
		move.l	#-1,0(a0,d0.l)	; Set expansion RAM
		tst.l	(a0)		; Still zero?
		bne	notexpansion
		move.l	0(a0,d0.l),d1
		addq.l	#1,d1
		bne	notexpansion
		move.l	#$100000,_obramsize	; Is 1 meg.
		move.l	a1,0(a0,d0.l)	; Restore memory.

notexpansion:
 D(8)
		move.l	a2,(a0)		; Restore memory
not512k:	move.l	#$10000,a7	; Working stack

 D(9)
		bsr	setupram	; Set up vectors, etc
 D(10)
		move.l	d7,-(a7)	; Flag to main
 D(11)
		jsr	_main		; To C code (does not return)

;
; Stick numbers into CRTC so that the 15 MHz RAM control starts
; A5 holds return address
;
		global	_crtc_tab
crtcstart:	move.l	#_crtc_tab,a0
		moveq.l	#0,d0		; Counter
wcrtc:
		move.b	d0,CRTCADDR
		move.b	(a0)+,CRTCDATA
		addq.l	#1,d0
		cmp.b	#14,d0
		bne.b	wcrtc
	; Wait for the internal registers to reload
		move.l	#100000,d0	; Over 20 msec
crwait:		subq.l	#1,d0		; 8
		bne.b	crwait		; 10. 18 * 100000 = (108 msec)
		jmp	(a5)

;
; Work out whether the CPU is 68000 or 68010
;

		global	_is68010
getcpu:		move.b	#1,_is68010
		move.l	ILLEGALVEC,d0	; Save current vector
		move.l	#ill,ILLEGALVEC
		move.l	sp,d1		; Save SP
		dc.w	$42c0		; MOVE CCR,d0 (68010 instr)
illret:		move.l	d1,sp		; Restore SP
		move.l	d0,ILLEGALVEC	; Restore vector
		rts
ill:		clr.b	_is68010
		bra.b	illret

;
; Set up all the hardware exception vectors, traps, etc
;

setupram:	move.l	#vtabe,d0
		sub.l	#vtab,d0	; Size of vector table
		move.l	#vtab,a0
L1:		move.l	(a0)+,a1	; Get pointer to vector address
		move.l	(a0)+,(a1)	; Move vector in
		subq.b	#8,d0
		bne.b	L1

		moveq.l	#15,d0		; Set up all trap vectors
		move.l	#$80,a0	; Pointer to trap vectors
L2:		move.l	#htrap,(a0)+
		dbf	d0,L2		; Do for 16 vectors

		rts

;
; Test for a bus error at the supplied location.
; Usage: berrtest(loc). Returns 1 if bus error, else 0
;

		global	_berrtest
_berrtest:
		move.l	4(a7),a0	; Location to test
		move.l	a2,-(sp)
		bsr	bt2
		move.l	(sp)+,a2
		rts

bt2:
		move.l	BERR_VEC,a1	; Save current vector
		move.l	#bthberr,BERR_VEC
		clr.l	d0
		move.l	sp,a2		; Save SP
		move.b	(a0),d1		; Maybe get a bus error: alters d0
gotberr:
		move.l	a1,BERR_VEC	; Restore vector
		rts

bthberr:
		move.l	a2,sp		; Restore sp
		moveq.l	#1,d0
		bra.b	gotberr

;
; Entry points for hardware exceptions, interrupts, etc
;

		global	_excepmes,_berrmes,_addresserr,_illegalmes,_zdividemes
		global	_chkmes,_trapvmes,_privmes,_l1111mes
		global	_unintmes,_spurmes,_trapmes,_irq1mes,_irq2mes,_irq3mes
		global	_irq4mes,_irq5mes,_irq6mes,_irq7mes
hbuserr:	move.l	#_berrmes,_excepmes
		bra	hexcep0
haddresserr:	move.l	#_addresserr,_excepmes
		bra	hexcep0
hillegal:	move.l	#_illegalmes,_excepmes
		bra	hexcep1
hzdivide:	move.l	#_zdividemes,_excepmes
		bra	hexcep2
hchk:		move.l	#_chkmes,_excepmes
		bra	hexcep2
htrapv:		move.l	#_trapvmes,_excepmes
		bra	hexcep2
hpriv:		move.l	#_privmes,_excepmes
		bra	hexcep1
;h1010:		move.l	#_l1010mes,_excepmes
;		bra	hexcep1
h1111:		move.l	#_l1111mes,_excepmes
		bra	hexcep1
hunint:		move.l	#_unintmes,_excepmes
		bra	hexcep1
hspur:		move.l	#_spurmes,_excepmes
		bra	hexcep1
htrap:		move.l	#_trapmes,_excepmes
		bra	hexcep2
hirq1:		move.l	#_irq1mes,_excepmes
		bra	hexcep1
hirq2:		move.l	#_irq2mes,_excepmes
		bra	hexcep1
hirq3:		move.l	#_irq3mes,_excepmes
		bra	hexcep1
hirq4:		move.l	#_irq4mes,_excepmes
		bra	hexcep1
hirq5:		move.l	#_irq5mes,_excepmes
		bra	hexcep1
hirq6:		move.l	#_irq6mes,_excepmes
		bra	hexcep1
hirq7:		move.l	#_irq7mes,_excepmes
		bra	hexcep1

		global	_trace
htrace:
		move.w	(a7),_srsave
		move.l	2(a7),_pcsave
		movem.l	d0/d1/d2/d3/d4/d5/d6/d7/a0/a1/a2/a3/a4/a5/a6/a7,_regsave
		jsr	_trace
		movem.l	_regsave,d0/d1/d2/d3/d4/d5/d6/d7/a0/a1/a2/a3/a4/a5/a6/a7
		rte

;
; Handle type 0 exceptions. Bus error and address error.
;
		global	_regsave,_fcsave,_irsave,_srsave,_wastype0
		global	_pcsave,_accaddr

hexcep0:
		tst.b	_is68010
		beq	hex068k

		clr.w	_fcsave			; 68010 exception 0 code
		move.l	10(a7),_accaddr
		move.l	24(a7),_irsave
		move.w	(a7),_srsave
		move.l	2(a7),_pcsave
		move.b	#1,_wastype0
		add.l	#58,a7
		bra.b	cont1
		
hex068k:	move.w	(a7)+,_fcsave		; 68000 exception 0 code
		move.l	(a7)+,_accaddr
		move.w	(a7)+,_irsave
		move.b	#1,_wastype0		; Flag exception type to C
		bra.b	contexcep

hexcep1:
hexcep2:	global	_cexception,_regsave
		clr.w	_wastype0
contexcep:	move.w	(a7)+,_srsave
		move.l	(a7)+,_pcsave
cont1:		movem.l	d0/d1/d2/d3/d4/d5/d6/d7/a0/a1/a2/a3/a4/a5/a6/a7,_regsave
		or.w	#$0700,sr
		move.w	_lastsyscall,_excepsyscall
		clr.l	-(sp)
		jsr	_cexception		; To C code

; Bad system call entry point
		global	_badsyscall,__empty
__empty:	move.l	#_badsyscall,_excepmes
		bra.b	cont1

; Table of vectors and their addresses
		global	__linea			; Line A handler
vtab:		dc.l	8,hbuserr
		dc.l	$c,haddresserr
		dc.l	$10,hillegal
		dc.l	$14,hzdivide
		dc.l	$18,hchk
		dc.l	$1c,htrapv
		dc.l	$20,hpriv
		dc.l	$24,htrace
		dc.l	$28,__linea
		dc.l	$2c,h1111
		dc.l	$3c,hunint
		dc.l	$60,hspur
		dc.l	$64,hirq1
		dc.l	$68,hirq2
		dc.l	$6c,hirq3
		dc.l	$70,hirq4
		dc.l	$74,hirq5
		dc.l	$78,hirq6
		dc.l	$7c,hirq7
vtabe:		nop

