# 1 "rommc.as"
| TABS8
		.text
|
| Applix 1616/OS ROM assembler startup code
|
| Copyright (C) 1987 by Applix pty limited
|
| X 28 Jan 1987. Programmer: Andrew K.P. Morton
|

	.set	BERR_VEC,8		| Bus error vector
	.set	ILLEGALVEC,16		| Illegal instruction vector

	.set	ROMSTART,0x500000	| Start of ROMs

	.set	CENTLATCH,0x600001	| Centronics latch
	.set	DACLATCH,0x600081	| DAC latch
	.set	VIDLATCH,0x600101	| Video latch
	.set	AMUXLATCH,0x600181	| Analog multiplexor latch
	.set	PAL0,0x600000	| Pallette entries
	.set	PAL1,0x600020
	.set	PAL2,0x600040
	.set	PAL3,0x600060

	.set	SCCBASE,0x700000	| SCC start
	.set	IPORT,0x700081	| Input port
	.set	VIABASE,0x700100	| VIA start
	.set	CRTCBASE,0x700180	| CRTC start

	.set	SCCBC,SCCBASE+0	| SCC B control
	.set	SCCBD,SCCBASE+2	| SCC B data
	.set	SCCAC,SCCBASE+4	| SCC A control
	.set	SCCAD,SCCBASE+6	| SCC A data

	.set	CRTCADDR,CRTCBASE	| CRTC address register
	.set	CRTCDATA,CRTCBASE+2	| CRTC data register

	.set	V_BREG,VIABASE		| Port B I/O register	*/
	.set	V_AREG,VIABASE+2	| Port A I/O register	*/
	.set	V_DDRB,VIABASE+4	| Port B DDR		*/
	.set	V_DDRA,VIABASE+6	| Port A DDR		*/

|
| Start of code
|

		.globl	_initsp,_main,__coldboot,__warmboot,_obramstart
		.globl	_pwruplongs,_asbootup,_excepsyscall
		.globl	_lastsyscall,_obramsize
		.globl	_user_patch




	.globl	_Tloc
	.comm	_Tloc,4











|__warmboot:
_asbootup:
bootup:		bra	tootle
_user_patch:	.long	0		| This is for user-patched ROMs
tootle:

	move.l	#0x11000,_Tloc

		move.b	d0,ROMSTART	| Bank out ROMs
		or.w	#0x0700,sr	| spl7
		move.l	_initsp,a7	| Set up Stack pointer (non-reset entry)
		reset
move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#1,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
dofn:
		move.b	IPORT,d0	| If switch 3 open
		bmi	toc		| set things up and branch to C

		move.b	d0,DACLATCH	| Echo it
		lsr.b	#4,d0		| Ignore bottom 4 bits
		and.l	#7,d0		| Only use 3 bits
		asl.l	#2,d0
		move.l	#ijtab,a0
		move.l	0(a0,d0.w),a0	| Get code address
		jmp	(a0)		| Go to it

__coldboot:	clr.l	_pwruplongs	| Ensure level 0 reset: zap the power up words
		jmp	__warmboot

ijtab:		.long	latchtest	| X 0: Test analog, video, dac, cent, pallette
		.long	sccrwtest	| X 1: SCC
		.long	viarwtest	| X 2: VIA
		.long	crtcrwtest	| X 3: CRTC
		.long	iporttest	| X 4: IPORT
		.long	memrwtest	| X 5: Read / write memory
		.long	memtest		| X 6: Memory test
		.long	vidtest		| X 7: Video test

| Write incrementing pattern to latches, pallette

latchtest:	clr.b	d3		| LED / relay bits
lt3:		move.l	#30000,d1	| Loop counter

lt2:		move.b	d0,CENTLATCH
		move.b	d0,DACLATCH
		move.b	d0,VIDLATCH

		move.b	d0,d2
		and.b	#0x77,d2	| Preserve LED & relay bits
		or.b	d3,d2
		move.b	d2,AMUXLATCH

		move.b	d0,PAL0
		move.b	d0,PAL1
		move.b	d0,PAL2
		move.b	d0,PAL3
		addq.b	#1,d0
		subq.l	#1,d1
		bne	lt2
		eor.b	#0x88,d3
		bra	lt3

| Read/write SCC

sccrwtest:	move.l	#SCCBASE,a0
		bra	rwtest

| Read/write VIA

viarwtest:
		move.b	#0xfe,V_DDRA		| Port A has one input bit.
		move.b	#0xff,V_DDRB
vrwtest:	clr.b	d2
vrwtest3:	move.l	#100000,d1
vrwtest2:	move.b	d0,V_AREG
		move.b	d0,V_BREG
		move.b	V_BREG,DACLATCH		| Echo the read back byte
		addq.b	#1,d0
		subq.l	#1,d1
		bne	vrwtest2
		eor.b	#0x8,d2
		move.b	d2,AMUXLATCH
		bra	vrwtest3

| Read/write CRTC

crtcrwtest:	move.l	#CRTCBASE,a0
		bra	rwtest

| Read from Input Port

iporttest:	move.l	#IPORT,a0
		bra	rtest

| Read / write memory

memrwtest:	move.l	#0x55aa,a0
		bra	rwtest

| General input port test: read from port and write to Centronics latch.
| Input port address in A0.

rtest:		clr.b	d1
rtest3:		move.l	#100000,d0
rtest2:		move.b	(a0),DACLATCH
		subq.l	#1,d0
		bne	rtest2
		eor.b	#0x8,d1
		move.b	d1,AMUXLATCH		| Flash LED
		bra	rtest3

| General output port test: increment latch contents. Latch address in A0

wtest:		clr.b	d2
wtest3:		move.l	#100000,d1
wtest2:		move.b	d0,(a0)
		addq.b	#1,d0
		subq.l	#1,d1
		bne	wtest2
		eor.b	#0x8,d2
		move.b	d2,AMUXLATCH
		bra	wtest3

|General read/write test. Write a value to port, read it back, echo to centronics port,
|increment value & loop

rwtest:		clr.b	d2
rwtest3:	move.l	#100000,d1
rwtest2:	move.b	d0,(a0)
		move.b	(a0),DACLATCH
		addq.b	#1,d0
		subq.l	#1,d1
		bne	rwtest2
		eor.b	#0x8,d2
		move.b	d2,AMUXLATCH
		bra	rwtest3

| Do memory test (this simplistic test will shortly be replaced)

memtest:	clr.w	d0
		clr.w	d1
		clr.w	d2		| Start value

memtest2:	move.w	d2,d0
		move.l	#0,a0
fillmem:	move.w	d0,(a0)+
		addq.w	#1,d0
		cmp.l	#0x80000,a0
		bne	fillmem

		move.w	d2,d0
		move.l	#0,a0
checkmem:	cmp.w	(a0)+,d0
		bne	badmem
		addq.w	#1,d0
		cmp.l	#0x80000,a0
		bne	checkmem

		add.w	#0x101,d2	| New pattern	
		eor.b	#0x88,d1
		move.b	d1,AMUXLATCH
		bra	memtest2

badmem:		subq.l	#2,a0		| Point to bad word
		move.l	a0,d0
		move.b	d0,DACLATCH	| Low byte of address
		lsr.l	#8,d0
		move.b	d0,VIDLATCH	| Next byte of address
		lsr.l	#8,d0
readbad2:	move.b	d0,AMUXLATCH	| Bits 16-23
		move.l	#30000,d1
readbad:	move.w	(a0),d2
		subq.l	#1,d1
		bne	readbad
		eor.b	#8,d0
		bra	readbad2

|
| Do Video test
|
		.globl	_cvidtest
vidtest:
		moveq.l	#1,d7
		bra	toc2

|vidtest:	move.l	#vtnext,a5
|		bra	crtcstart
|vtnext:
|		move.l	#0x40000,a7	| Middle of RAM
|		bsr	getcpu		|Ascertain CPU type
|		bsr	setupram
|		clr.l	_obramst:w
|		jmp	_cvidtest

|
| Code executed when switch 3 not in test mode
|

toc:		clr.l	d7
toc2:		move.l	#tocnext,a5
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#2,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		bra	crtcstart
tocnext:
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#3,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		bsr	getcpu		|Ascertain CPU type
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#4,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		clr.l	d3		| Search for start of on-board RAM
btloop:		add.l	#0x00100000,d3	| Go in 1 Meg steps
		move.l	 d3,-(a7)
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#5,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		bsr	_berrtest	| See if there is RAM there
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#6,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		addq.l	#4,a7		| Adjust stack
		tst.w	d0		| Returns 1 if bus error
		bne	noram
		cmp.l	#0x00500000,d3	| Do we have a full complement?
		bne	btloop
noram:		sub.l	#0x00100000,d3	| Start of on-board RAM
		move.l	d3,_obramstart
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#7,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
| New stuff for 1/2 meg expansion
		move.l	#0x80000,d0
		move.l	d0,_obramsize	| Assume 1/2 meg only.
		move.l	d3,a0		| Point to start of on-board RAM
		move.l	(a0),a2		| Save what's there
		clr.l	(a0)
		move.l	0(a0,d0.l),a1	| Save what's there
		move.l	#-1,0(a0,d0.l)	| Set expansion RAM
		tst.l	(a0)		| Still zero?
		bne	notexpansion
		move.l	0(a0,d0.l),d1
		addq.l	#1,d1
		bne	notexpansion
		move.l	#0x100000,_obramsize	| Is 1 meg.
		move.l	a1,0(a0,d0.l)	| Restore memory.

notexpansion:
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#8,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		move.l	a2,(a0)		| Restore memory
not512k:	move.l	#0x10000,a7	| Working stack

 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#9,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		bsr	setupram	| Set up vectors, etc
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#10,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		move.l	d7,-(a7)	| Flag to main
 move.l	a0,0x4100	
	move.l	_Tloc,a0	
	move.l	#11,(a0)+	
	move.l	a0,_Tloc	
	move.l	0x4100,a0
		jbsr	_main		| To C code (does not return)

|
| Stick numbers into CRTC so that the 15 MHz RAM control starts
| A5 holds return address
|
		.globl	_crtc_tab
crtcstart:	move.l	#_crtc_tab,a0
		moveq.l	#0,d0		| Counter
wcrtc:
		move.b	d0,CRTCADDR
		move.b	(a0)+,CRTCDATA
		addq.l	#1,d0
		cmp.b	#14,d0
		bne	wcrtc
	| Wait for the internal registers to reload
		move.l	#100000,d0	| Over 20 msec
crwait:		subq.l	#1,d0		| X 8
		bne	crwait		| X 10. 18 * 100000 = (108 msec)
		jmp	(a5)

|
| Work out whether the CPU is 68000 or 68010
|

		.globl	_is68010
getcpu:		move.b	#1,_is68010
		move.l	ILLEGALVEC,d0	| Save current vector
		move.l	#ill,ILLEGALVEC
		move.l	sp,d1		| Save SP
		.word	0x42c0		| MOVE CCR,d0 (68010 instr)
illret:		move.l	d1,sp		| Restore SP
		move.l	d0,ILLEGALVEC	| Restore vector
		rts
ill:		clr.b	_is68010
		bra	illret

|
| Set up all the hardware exception vectors, traps, etc
|

setupram:	move.l	#vtabe,d0
		sub.l	#vtab,d0	| Size of vector table
		move.l	#vtab,a0
L1:		move.l	(a0)+,a1	| Get pointer to vector address
		move.l	(a0)+,(a1)	| Move vector in
		subq.b	#8,d0
		bne	L1

		moveq.l	#15,d0		| Set up all trap vectors
		move.l	#0x80,a0	| Pointer to trap vectors
L2:		move.l	#htrap,(a0)+
		dbf	d0,L2		| Do for 16 vectors

		rts

|
| Test for a bus error at the supplied location.
| Usage: berrtest(loc). Returns 1 if bus error, else 0
|

		.globl	_berrtest
_berrtest:
		move.l	4(a7),a0	| Location to test
		move.l	a2,-(sp)
		bsr	bt2
		move.l	(sp)+,a2
		rts

bt2:
		move.l	BERR_VEC,a1	| Save current vector
		move.l	#bthberr,BERR_VEC
		clr.l	d0
		move.l	sp,a2		| Save SP
		move.b	(a0),d1		| Maybe get a bus error: alters d0
gotberr:
		move.l	a1,BERR_VEC	| Restore vector
		rts

bthberr:
		move.l	a2,sp		| Restore sp
		moveq.l	#1,d0
		bra	gotberr

|
| Entry points for hardware exceptions, interrupts, etc
|

		.globl	_excepmes,_berrmes,_addresserr,_illegalmes,_zdividemes
		.globl	_chkmes,_trapvmes,_privmes,_l1111mes
		.globl	_unintmes,_spurmes,_trapmes,_irq1mes,_irq2mes,_irq3mes
		.globl	_irq4mes,_irq5mes,_irq6mes,_irq7mes
hbuserr:	move.l	#_berrmes,_excepmes
		bra	hexcep0
haddresserr:	move.l	#_addresserr,_excepmes
		bra	hexcep0
hillegal:	move.l	#_illegalmes,_excepmes
		bra	hexcep1
hzdivide:	move.l	#_zdividemes,_excepmes
		bra	hexcep2
hchk:		move.l	#_chkmes,_excepmes
		bra	hexcep2
htrapv:		move.l	#_trapvmes,_excepmes
		bra	hexcep2
hpriv:		move.l	#_privmes,_excepmes
		bra	hexcep1
|h1010:		move.l	#_l1010mes,_excepmes
|		bra	hexcep1
h1111:		move.l	#_l1111mes,_excepmes
		bra	hexcep1
hunint:		move.l	#_unintmes,_excepmes
		bra	hexcep1
hspur:		move.l	#_spurmes,_excepmes
		bra	hexcep1
htrap:		move.l	#_trapmes,_excepmes
		bra	hexcep2
hirq1:		move.l	#_irq1mes,_excepmes
		bra	hexcep1
hirq2:		move.l	#_irq2mes,_excepmes
		bra	hexcep1
hirq3:		move.l	#_irq3mes,_excepmes
		bra	hexcep1
hirq4:		move.l	#_irq4mes,_excepmes
		bra	hexcep1
hirq5:		move.l	#_irq5mes,_excepmes
		bra	hexcep1
hirq6:		move.l	#_irq6mes,_excepmes
		bra	hexcep1
hirq7:		move.l	#_irq7mes,_excepmes
		bra	hexcep1

		.globl	_trace
htrace:
		move.w	(a7),_srsave
		move.l	2(a7),_pcsave
		movem.l	d0/d1/d2/d3/d4/d5/d6/d7/a0/a1/a2/a3/a4/a5/a6/a7,_regsave
		jbsr	_trace
		movem.l	_regsave,d0/d1/d2/d3/d4/d5/d6/d7/a0/a1/a2/a3/a4/a5/a6/a7
		rte

|
| Handle type 0 exceptions. Bus error and address error.
|
		.globl	_regsave,_fcsave,_irsave,_srsave,_wastype0
		.globl	_pcsave,_accaddr

hexcep0:
		tst.b	_is68010
		beq	hex068k

		clr.w	_fcsave			| X 68010 exception 0 code
		move.l	10(a7),_accaddr
		move.l	24(a7),_irsave
		move.w	(a7),_srsave
		move.l	2(a7),_pcsave
		move.b	#1,_wastype0
		add.l	#58,a7
		bra	cont1
		
hex068k:	move.w	(a7)+,_fcsave		| X 68000 exception 0 code
		move.l	(a7)+,_accaddr
		move.w	(a7)+,_irsave
		move.b	#1,_wastype0		| Flag exception type to C
		bra	contexcep

hexcep1:
hexcep2:	.globl	_cexception,_regsave
		clr.w	_wastype0
contexcep:	move.w	(a7)+,_srsave
		move.l	(a7)+,_pcsave
cont1:		movem.l	d0/d1/d2/d3/d4/d5/d6/d7/a0/a1/a2/a3/a4/a5/a6/a7,_regsave
		or.w	#0x0700,sr
		move.w	_lastsyscall,_excepsyscall
		clr.l	-(sp)
		jbsr	_cexception		| To C code

| Bad system call entry point
		.globl	_badsyscall,__empty
__empty:	move.l	#_badsyscall,_excepmes
		bra	cont1

| Table of vectors and their addresses
		.globl	__linea			| Line A handler
vtab:		.long	8,hbuserr
		.long	0xc,haddresserr
		.long	0x10,hillegal
		.long	0x14,hzdivide
		.long	0x18,hchk
		.long	0x1c,htrapv
		.long	0x20,hpriv
		.long	0x24,htrace
		.long	0x28,__linea
		.long	0x2c,h1111
		.long	0x3c,hunint
		.long	0x60,hspur
		.long	0x64,hirq1
		.long	0x68,hirq2
		.long	0x6c,hirq3
		.long	0x70,hirq4
		.long	0x74,hirq5
		.long	0x78,hirq6
		.long	0x7c,hirq7
vtabe:		nop

