/* TABS4
 */

#ifndef CANFDX
#define ZMODEM_MODULE
#include "zmodem.h"
#endif

#include <syscalls.h>

#define noxrd7() readline(Rxtimeout)


long		Bytesleft;			/* number of bytes of incoming file left */
long		AKPMtotsize;
int			Zfile_fd;
unsigned	Baudrate;
int			Zlastsent;

int			Rxframeind;			/* ZBIN ZBIN32, or ZHEX type of frame received */
int			Rxtype;				/* Type of header received */
int			Rxcount;			/* Count of data bytes received */
int			Rxtimeout;			/* Tenths of seconds to wait for something */
char		Rxhdr4;			/* Received header */
char		Txhdr4;			/* Transmitted header */
long		Rxpos;				/* Received file position */
long		Txpos;				/* Transmitted file position */
char		AttnZATTNLEN + 1;	/* Attention string rx sends to tx on err */

extern char Zisinteractive;
extern char *ZPathname;

/* Send ZMODEM binary header hdr of type type */
zsbhdr(type, hdr)
register char *hdr;
{
	register n;
	register unsigned short crc;

	sendline(ZPAD);
	sendline(ZDLE);
	sendline(ZBIN);
	zsendline(type);
	crc = updcrc(type, 0);

	for (n = 4; --n >= 0;)
	{
		zsendline(*hdr);
		crc = updcrc((0377 & *hdr++), crc);
	}
	crc = updcrc(0, updcrc(0, crc));
	zsendline(crc >> 8);
	zsendline(crc);
	sl_bflush();
}

#if 0
/* Send ZMODEM binary header hdr of type type */
zsbh32(hdr, type)
register char *hdr;
{
	register n;
	register unsigned long crc;

	sendline(ZBIN32);
	zsendline(type);
	crc = 0xFFFFFFFF;
	crc = UPDC32(type, crc);

	for (n = 4; --n >= 0;)
	{
		zsendline(*hdr);
		crc = UPDC32((0377 & *hdr++), crc);
	}
	crc = ~crc;
	for (n = 4; --n >= 0;)
	{
		zsendline(crc);
		crc >>= 8;
	}
	sl_bflush();
}

#endif

/* Send ZMODEM HEX header hdr of type type */
zshhdr(type, hdr)
register char *hdr;
{
	register n;
	register unsigned short crc;

	sendline(ZPAD);
	sendline(ZPAD);
	sendline(ZDLE);
	sendline(ZHEX);
	zputhex(type);

	crc = updcrc(type, 0);
	for (n = 4; --n >= 0;)
	{
		zputhex(*hdr);
		crc = updcrc((0377 & *hdr++), crc);
	}
	crc = updcrc(0, updcrc(0, crc));
	zputhex(crc >> 8);
	zputhex(crc);

	/* Make it printable on remote machine */
	sendline(015);
	sendline(012);

	/*
	 * Uncork the remote in case a fake XOFF has stopped data flow
	 */
	if (type != ZFIN)
		sendline(021);
	sl_bflush();
}

/*
 * Send binary array buf of length length, with ending ZDLE sequence frameend
 */
zsdata(buf, length, frameend)
register char *buf;
{
	register unsigned short crc;
	int tellpos = TELL(Zfile_fd);

	if (Zisinteractive)
	{
		AMprintf("\001Sending '%s': position=%d (%02d%%)        ",
			ZPathname, tellpos,
			(100 * tellpos) / (AKPMtotsize ? AKPMtotsize : 100000));
	}

	crc = 0;
	while (length-- > 0)
	{
		zsendline(*buf);
		crc = updcrc((0377 & *buf++), crc);
	}
	sendline(ZDLE);
	sendline(frameend);
	crc = updcrc(frameend, crc);

	crc = updcrc(0, updcrc(0, crc));
	zsendline(crc >> 8);
	zsendline(crc);
	if (frameend == ZCRCW)
		sendline(XON);
	sl_bflush();
}

#if 0
zsda32(buf, length, frameend)
register char *buf;
{
	register unsigned long crc;

	crc = 0xFFFFFFFF;
	for (; --length >= 0;)
	{
		zsendline(*buf);
		crc = UPDC32((0377 & *buf++), crc);
	}
	sendline(ZDLE);
	sendline(frameend);
	crc = UPDC32(frameend, crc);

	crc = ~crc;
	for (length = 4; --length >= 0;)
	{
		zsendline(crc);
		crc >>= 8;
	}
	sl_bflush();
}

#endif

Rxstatus()
{
	if (Zisinteractive && ZPathname0)
	{
		AMprintf("\001Receiving '%s': position=%d (%02d%%)            ",
			 ZPathname,
			 AKPMtotsize - Bytesleft,
			 ((AKPMtotsize - Bytesleft) * 100) /
			 (AKPMtotsize ? AKPMtotsize : 100000));
	}
}

/*
 * Receive array buf of max length with ending ZDLE sequence
 *  and CRC.  Returns the ending character or error code.
 */
zrdata(buf, length)
register char *buf;
{
	register c;
	register unsigned short crc;
	register d;

	if (Rxframeind == ZBIN32)
		return zrdat32(buf, length);

	crc = Rxcount = 0;
	for (;;)
	{
		if ((c = zdlread()) & ~0377)
		{
	crcfoo:
			switch (c)
			{
			case GOTCRCE:
			case GOTCRCG:
			case GOTCRCQ:
			case GOTCRCW:
				crc = updcrc((d = c) & 0377, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = updcrc(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = updcrc(c, crc);
				if (crc & 0xFFFF)
				{
					AMprintf("Bad data CRC %x", crc);
					return ERROR;
				}
				Rxstatus();
				Bytesleft -= Rxcount;
				return d;
			case GOTCAN:
				AMprintf("ZMODEM: Sender Canceled");
				return ZCAN;
			case TIMEOUT:
				AMprintf("ZMODEM data TIMEOUT");
				return c;
			default:
				AMprintf("ZMODEM bad data subpacket ret=%x", c);
				return c;
			}
		}
		if (--length < 0)
		{
			AMprintf("ZMODEM data subpacket too long");
			return ERROR;
		}
		++Rxcount;
		*buf++ = c;
		crc = updcrc(c, crc);
		continue;
	}
}

zrdat32(buf, length)
register char *buf;
{
	register c;
	register unsigned long crc;
	register d;

	crc = 0xFFFFFFFF;
	Rxcount = 0;
	for (;;)
	{
		if ((c = zdlread()) & ~0377)
		{
	crcfoo:
			switch (c)
			{
			case GOTCRCE:
			case GOTCRCG:
			case GOTCRCQ:
			case GOTCRCW:
				crc = UPDC32((d = c) & 0377, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if ((c = zdlread()) & ~0377)
					goto crcfoo;
				crc = UPDC32(c, crc);
				if (crc != 0xDEBB20E3)
				{
					AMprintf("Bad data CRC %x", crc);
					return ERROR;
				}
				Rxstatus();
				Bytesleft -= Rxcount;
				return d;
			case GOTCAN:
				AMprintf("ZMODEM: Sender Canceled");
				return ZCAN;
			case TIMEOUT:
				AMprintf("ZMODEM data TIMEOUT");
				return c;
			default:
				AMprintf("ZMODEM bad data subpacket ret=%x", c);
				return c;
			}
		}
		if (--length < 0)
		{
			AMprintf("ZMODEM data subpacket too long");
			return ERROR;
		}
		++Rxcount;
		*buf++ = c;
		crc = UPDC32(c, crc);
		continue;
	}
}

/*
 * Read a ZMODEM header to hdr, either binary or hex.
 *  eflag controls local display of non zmodem characters:
 *	0:  no display
 *	1:  display printing characters only
 *	2:  display all non ZMODEM characters
 *  On success, set Zmodem to 1 and return type of header.
 *   Otherwise return negative on error
 */
zgethdr(hdr)
char *hdr;
{
	register c, n, cancount;

	n = Baudrate;		/* Max characters before start of frame */
	cancount = 5;
again:
	Rxframeind = Rxtype = 0;
	switch (c = noxrd7())
	{
	case RCDO:
	case TIMEOUT:
		goto fifi;
	case CAN:
		if (--cancount <= 0)
		{
			c = ZCAN;
			goto fifi;
		}
		/* **** FALL THRU TO **** */
	default:
agn2:
		if (--n == 0)
		{
			AMprintf("ZMODEM Garbage count exceeded");
			return (ERROR);
		}
		if (c != CAN)
			cancount = 5;
		goto again;
	case ZPAD:		/* This is what we want. */
		break;
	}
	cancount = 5;
splat:
	switch (c = noxrd7())
	{
	case ZPAD:
		goto splat;
	case RCDO:
	case TIMEOUT:
		goto fifi;
	default:
		goto agn2;
	case ZDLE:		/* This is what we want. */
		break;
	}

	switch (c = noxrd7())
	{
	case RCDO:
	case TIMEOUT:
		goto fifi;
	case ZBIN:
		Rxframeind = ZBIN;
		c = zrbhdr(hdr);
		break;
	case ZBIN32:
		Rxframeind = ZBIN32;
		c = zrbhdr32(hdr);
		break;
	case ZHEX:
		Rxframeind = ZHEX;
		c = zrhhdr(hdr);
		break;
	case CAN:
		if (--cancount <= 0)
		{
			c = ZCAN;
			goto fifi;
		}
		goto agn2;
	default:
		goto agn2;
	}
	Rxpos = hdrZP3 & 0377;
	Rxpos = (Rxpos << 8) + (hdrZP2 & 0377);
	Rxpos = (Rxpos << 8) + (hdrZP1 & 0377);
	Rxpos = (Rxpos << 8) + (hdrZP0 & 0377);
fifi:
	switch (c)
	{
	case GOTCAN:
		c = ZCAN;
		/* **** FALL THRU TO **** */
	case ZNAK:
	case ZCAN:
	case ERROR:
	case TIMEOUT:
	case RCDO:
#if 0
		AMprintf("ZMODEM: Got %s %s",
			 frametypesc + FTOFFSET,
			 (c >= 0) ? "header" : "error");
#endif
		;
	}
	return c;
}

/* Receive a binary style header (type and position) */
zrbhdr(hdr)
register char *hdr;
{
	register c, n;
	register unsigned short crc;

	if ((c = zdlread()) & ~0377)
		return c;
	Rxtype = c;
	crc = updcrc(c, 0);

	for (n = 4; --n >= 0;)
	{
		if ((c = zdlread()) & ~0377)
			return c;
		crc = updcrc(c, crc);
		*hdr++ = c;
	}
	if ((c = zdlread()) & ~0377)
		return c;
	crc = updcrc(c, crc);
	if ((c = zdlread()) & ~0377)
		return c;
	crc = updcrc(c, crc);
	if (crc & 0xFFFF)
	{
		AMprintf("Bad Header CRC");
		return ERROR;
	}
	return Rxtype;
}

/* Receive a binary style header (type and position) with 32 bit FCS */
zrbhdr32(hdr)
register char *hdr;
{
	register c, n;
	register unsigned long crc;

	if ((c = zdlread()) & ~0377)
		return c;
	Rxtype = c;
	crc = 0xFFFFFFFF;
	crc = UPDC32(c, crc);

	for (n = 4; --n >= 0;)
	{
		if ((c = zdlread()) & ~0377)
			return c;
		crc = UPDC32(c, crc);
		*hdr++ = c;
	}
	for (n = 4; --n >= 0;)
	{
		if ((c = zdlread()) & ~0377)
			return c;
		crc = UPDC32(c, crc);
	}
	if (crc != 0xDEBB20E3)
	{
		AMprintf("Bad Header CRC %x", crc);
		return ERROR;
	}
	return Rxtype;
}

/* Receive a hex style header (type and position) */
zrhhdr(hdr)
char *hdr;
{
	register c;
	register unsigned short crc;
	register n;

	if ((c = zgethex()) < 0)
		return c;
	Rxtype = c;
	crc = updcrc(c, 0);

	for (n = 4; --n >= 0;)
	{
		if ((c = zgethex()) < 0)
			return c;
		crc = updcrc(c, crc);
		*hdr++ = c;
	}
	if ((c = zgethex()) < 0)
		return c;
	crc = updcrc(c, crc);
	if ((c = zgethex()) < 0)
		return c;
	crc = updcrc(c, crc);
	if (crc & 0xFFFF)
	{
		AMprintf("Bad Header CRC");
		return ERROR;
	}
	if (readline(1) == '\r')/* Throw away possible cr/lf */
		readline(1);
	return Rxtype;
}

/* Send a byte as two hex digits */
zputhex(c)
register c;
{
	static char digits = "0123456789abcdef";

	sendline(digits(c & 0xF0) >> 4);
	sendline(digits(c) & 0xF);
	sl_bflush();
}

/*
 * Send character c with ZMODEM escape sequence encoding.
 *  Escape XON, XOFF. Escape CR following @ (Telenet net escape)
 */
zsendline(c)
unsigned char c;
{
	switch (c)
	{
	case ZDLE:
		sendline(ZDLE);
		sendline(Zlastsent = (c ^= 0100));
		break;
	case 015:
	case 0215:
		if ((Zlastsent & 0177) != '@')
			goto sendit;
		/* **** FALL THRU TO **** */
	case 020:
	case 021:
	case 023:
	case 0220:
	case 0221:
	case 0223:
		sendline(ZDLE);
		c ^= 0100;
sendit:
	default:
		sendline(Zlastsent = c);
		break;
	}
}

/* Decode two lower case hex digits into an 8 bit byte value */
zgethex()
{
	register c;

	c = zgeth1();
	return c;
}

zgeth1()
{
	register c, n;

	if ((c = noxrd7()) < 0)
		return c;
	n = c - '0';
	if (n > 9)
		n -= ('a' - ':');
	if (n & ~0xF)
		return ERROR;
	if ((c = noxrd7()) < 0)
		return c;
	c -= '0';
	if (c > 9)
		c -= ('a' - ':');
	if (c & ~0xF)
		return ERROR;
	c += (n << 4);
	return c;
}

/*
 * Read a byte, checking for ZMODEM escape encoding
 *  including CAN*5 which represents a quick abort
 */
zdlread()
{
	register c;

	if ((c = readline(Rxtimeout)) != ZDLE)
		return c;
	if ((c = readline(Rxtimeout)) < 0)
		return c;
	if (c == CAN && (c = readline(Rxtimeout)) < 0)
		return c;
	if (c == CAN && (c = readline(Rxtimeout)) < 0)
		return c;
	if (c == CAN && (c = readline(Rxtimeout)) < 0)
		return c;
	switch (c)
	{
	case CAN:
		return GOTCAN;
	case ZCRCE:
	case ZCRCG:
	case ZCRCQ:
	case ZCRCW:
		return (c | GOTOR);
	case ZRUB0:
		return 0177;
	case ZRUB1:
		return 0377;
	default:
		if ((c & 0140) == 0100)
			return (c ^ 0100);
		break;
	}
	AMprintf("Got bad ZMODEM escape sequence %x", c);
	return ERROR;
}

/* Store long integer pos in Txhdr */
stohdr(pos)
long pos;
{
	TxhdrZP0 = pos;
	TxhdrZP1 = pos >> 8;
	TxhdrZP2 = pos >> 16;
	TxhdrZP3 = pos >> 24;
}

/* Recover a long integer from a header */
long
rclhdr(hdr)
register unsigned char *hdr;
{
	register unsigned long l;

	l = hdrZP3;
	l = (l << 8) | hdrZP2;
	l = (l << 8) | hdrZP1;
	l = (l << 8) | hdrZP0;
	return (long) l;
}
